

/*	+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	-----------------------------------------------------------------
	|	SECTION 2:													|
	|		TUTORIAL 2.1: Creating an ecore-Model using C++-Code	|
	-----------------------------------------------------------------
	+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

//Incude standard libraries
#include <iostream>

//Include header files for ecore
#include "abstractDataTypes/SubsetUnion.hpp"
#include "ecore/EcoreFactory.hpp"
#include "ecore/EcorePackage.hpp"
#include "ecore/EAttribute.hpp"
#include "ecore/EClass.hpp"
#include "ecore/EClassifier.hpp"
#include "ecore/EReference.hpp"
#include "ecore/EDataType.hpp"
#include "ecore/EStringToStringMapEntry.hpp"
#include "ecore/EStructuralFeature.hpp"
#include "types/TypesPackage.hpp"

using namespace ecore;
using namespace types;

int main()
{
 //Create an instance of an ecore package and an ecore factory (both are needed to create an ecore model)
  std::shared_ptr<EcorePackage> ecorePackage = EcorePackage::eInstance();
  std::shared_ptr<EcoreFactory> ecoreFactory = EcoreFactory::eInstance();

//Create instances of Ecore-Elemets


	//------------------------------------------------------------------
	//2.1.2	Creating inctances of ecore-Elements using the ecore factory
	//------------------------------------------------------------------
		// Create an instance of EPackage (using the ecore factory)
		std::shared_ptr<EPackage> rootPackage = ecoreFactory->createEPackage(); // root-Package has no SuperPackage
		rootPackage->setName("UniExample_ecore");
		
		// Create an instance of EClass (which will become our UniModel classes) in root package that we just created (using ecore factory)
		std::shared_ptr<EClass> eClass_student = ecoreFactory->createEClass_in_EPackage(rootPackage);
		
		// Create an instance of EAttribute for the EClass that we just created (using the ecore factory)
		std::shared_ptr<EAttribute> eAttribute_student_matNum = ecoreFactory->createEAttribute_in_EContainingClass(eClass_student);

		
		//Create an instance of TypesPackage
		std::shared_ptr<TypesPackage> typesPackage = TypesPackage::eInstance();
		//Assign data type "Integer" to the matriculation number EAttribute
		eAttribute_student_matNum->setEType(typesPackage->getInteger_Class());
		
		// Set properties of our ecore model elements 
		eClass_student->setName("Student");
		eAttribute_student_matNum->setName("matNum");

  
  
	//----------------------------------------------------------------------------------
	//2.1.2 Create instances of ecore-Elements using the name of the elements' metaclass
	//----------------------------------------------------------------------------------
		// Creating an instance of EObject (which is the root object for all modeled elements) 
		// by using one of the ecore factorys' create()-methods which takes a string parameter --> a name of a meta class
		std::shared_ptr<EObject> eObject_1 = ecoreFactory->create("EClass");
		
		//Forward declaration
		std::shared_ptr<EClass> eClass_faculty;

		if (eObject_1 != nullptr) {
			// Cast eObject_1 to a new EClass instance
			eClass_faculty = std::dynamic_pointer_cast<EClass>(eObject_1);
			if (eClass_faculty != nullptr) {
				
				eClass_faculty->setName("Faculty");
				
				//Creating an EReference via EObject using metaclass name
				std::shared_ptr<EObject> eObject_2 = ecoreFactory->create("EReference");
				
				//Forward declaration
				std::shared_ptr<EReference> eReference_faculty_student;
				
				if(eObject_2 != nullptr){
					// Cast eObject_2 to a new EReference instance
					eReference_faculty_student = std::dynamic_pointer_cast<EReference>(eObject_2);
					
					if(eReference_faculty_student != nullptr){
						
						eReference_faculty_student->setName("students");
						
						//Reference will be of type "Student"
						eReference_faculty_student->setEType(eClass_student);
						
						//-1 is used to set a [0..*] multiplicity
						eReference_faculty_student->setUpperBound(-1);
						
						//Add our new reference to our lecture class
						//In MDE4CPP Bag<> is used as container class
						std::shared_ptr<Bag<ecore::EReference> > references = eClass_faculty->getEReferences();
						references->add(eReference_faculty_student);
					}
				}
			
				// Add our new class to our package
				//In MDE4CPP Bag<> is used as container class
				std::shared_ptr<Bag<ecore::EClassifier> > classifiers = rootPackage->getEClassifiers();
				classifiers->add(eClass_faculty);
			}
		}
	
	
	
	//--------------------------------------------------------------------------------
	//2.1.3 Create instances of ecore-Elements using the ID of the elements' metaclass
	//--------------------------------------------------------------------------------
		// Create an instance of EClass using the EObject->eClass() method (which returns the meta class information of eObject_1
		// Since eObject_1 was instanciated by meta class name "EClass" in 2.1.2 this will return meta class information for EClass
		std::shared_ptr<EClass> eClass_metaclass = eObject_1->eClass();
		
		//Create a new instance of EObject using the classifier ID of meta class EClass
		std::shared_ptr<EObject> eObject_3 = ecoreFactory->create(eClass_metaclass->getClassifierID());
		
		//Forward declaration
		std::shared_ptr<EClass> eClass_exam;
		
			if (eObject_3 != nullptr) {
				//Cast eObject_3 to a new EClass instance
				eClass_exam = std::dynamic_pointer_cast<EClass>(eObject_3);
				if (eClass_exam != nullptr) {
					
					eClass_exam->setName("Exam");
					
					// Create an additional attribute in our new class
					std::shared_ptr<EAttribute> eAttribute_exam_score = ecoreFactory->createEAttribute_in_EContainingClass(eClass_exam);
					eAttribute_exam_score->setName("score");
					//Assign data type "Real"(=Float) to the exam score EAttribute
					eAttribute_exam_score->setEType(typesPackage->getReal_Class());
					
					//Add an additional EClass
					std::shared_ptr<EClass> eClass_lecture = ecoreFactory->createEClass_in_EPackage(rootPackage);
					eClass_lecture->setName("Lecture");
					
					// Add an additional reference between Exam class and Lecture class
					std::shared_ptr<EReference> eReference_exam_lecture = ecoreFactory->createEReference_in_EContainingClass(eClass_exam);
					eReference_exam_lecture->setName("lecture");
					// Reference will be of type "Lecture"
					if(eClass_lecture != nullptr){eReference_exam_lecture->setEType(eClass_lecture);}
					
					//Lets add another additional reference between the "Faculty"-class and the "Lecture"-class
					std::shared_ptr<EReference> eReference_faculty_lecture = ecoreFactory->createEReference_in_EContainingClass(eClass_faculty);
					eReference_faculty_lecture->setName("lectures");
					eReference_faculty_lecture->setUpperBound(-1);
					// Reference will be of type "Lecture"
					if(eClass_student != nullptr){eReference_faculty_lecture->setEType(eClass_lecture);}
					
					// Add our new class to our package
					//In MDE4CPP Bag<> is used as container class
					std::shared_ptr<Bag<ecore::EClassifier> > classifiers = rootPackage->getEClassifiers();
					classifiers->add(eClass_exam);
				}
			}
			
			
			
  //------------------------------------
  //Output contents of our model package
  //------------------------------------
		std::cout<<std::endl<<"-" << rootPackage->getName() << " package content-"<<std::endl<<"************"<<std::endl;
		//Create container of all classifiers (in our case just EClass) of our package
		std::shared_ptr<Bag<ecore::EClassifier> > classifiers = rootPackage->getEClassifiers();
		
		Bag<ecore::EClassifier>::const_iterator endIt         = classifiers->end();
		//Loop over all classifiers of the package
		for (Bag<ecore::EClassifier>::const_iterator it = classifiers->begin(); it != endIt; ++it) {
			std::cout << "class name: " << (*it)->getName() << std::endl;
			
			//Cast iterator to EClass
			std::shared_ptr<EClass> currentClassifier = std::dynamic_pointer_cast<EClass>(*it);

			std::cout << "	Features:"<< std::endl;
			//Create container of all features (attributes, references, ...) of the current classifier
			std::shared_ptr<Bag<ecore::EStructuralFeature> > structuralFeatures = currentClassifier->getEAllStructuralFeatures();
			
			Bag<ecore::EStructuralFeature>::const_iterator endItFeature         = structuralFeatures->end();
			//Loop over all features of the current classifier
			for (Bag<ecore::EStructuralFeature>::const_iterator itFeature = structuralFeatures->begin(); itFeature != endItFeature; ++itFeature)
			{
				std::string featName = (*itFeature)->getName();
				// eClass() returns the meta class information of eClass_student, eClass()->getName() returns the name of the meta class
				std::string metaClassName = (*itFeature)->eClass()->getName();
				std::string typeName = "unknown";	//in case the current feature does not have a type
				
				int upperBound = (*itFeature)->getUpperBound();
				int lowerBound = (*itFeature)->getLowerBound();
				
				std::shared_ptr<ecore::EClassifier> featClassifier = (*itFeature)->getEType();
				if(featClassifier != nullptr) typeName = featClassifier->getName();
				
				std::cout << "		- '"<< featName << "': " << typeName << "[" << lowerBound << ".." << upperBound <<"]\t\tof meta class " << metaClassName << std::endl;
				
				
			}
			std::cout << "____________" << std::endl;
		}
  

		std::cout << "Finished..." << std::endl;

		return 0;
  
}
